from __future__ import annotations
import numpy as np
from numpy import ndarray
from typing import Protocol, Tuple, cast

from torch import wait
from xtensors.typing import NDArray
from xarray import DataArray


def are_shapes_broadcastable(shape1: Tuple[int,...], shape2: Tuple[int,...]) -> bool:
    for a, b in zip(shape1[::-1], shape2[::-1]):
        if a == 1 or b == 1 or a == b:
            pass
        else:
            return False
    return True


def broadcast_shapes(shape1: Tuple[int,...], shape2: Tuple[int,...]) -> Tuple[int,...]:
    if len(shape1) < len(shape2): return broadcast_shapes(shape2, shape1)

    if not are_shapes_broadcastable: raise IndexError(f'shapes {shape1} and {shape2} cannot be broadcast')

    shape = list(shape1)[::-1]

    for axis, s2 in enumerate(list(shape2)[::-1]):
        if s2 != 1: shape[axis] = s2

    return tuple(shape[::-1])


def are_arrays_broadcastable(x: NDArray, y: NDArray) -> bool:
    if isinstance(x, DataArray) and isinstance(y, DataArray): return are_xarrays_broadcastable(x, y)

    return are_shapes_broadcastable(x.shape, y.shape)


def are_xarrays_broadcastable(x: DataArray, y: DataArray) -> bool:
    '''
    The broadcastability of two xarray.DataArrays is limited to only special
    cases for stability reasons. Two DataArrays x and y are broadcastable if
    and only if:

        x.dims contain y.dims, and the corresponding axes are compatible

        or

        y .dims contain x.dims, and the corresponding axes are compatible


    '''
    dim_x = x.dims
    dim_y = y.dims
    
    if len(dim_x) >= len(dim_y):
        if all([dim in dim_x for dim in dim_y]):
            for dim in dim_y:
                axis_x = cast(Tuple[int], x.get_axis_num((dim,)))[0]
                axis_y = cast(Tuple[int], y.get_axis_num((dim,)))[0]
                if x.shape[axis_x] != y.shape[axis_y] and x.shape[axis_x] != 1 and y.shape[axis_y] != 1:
                    return False
            return True
        return False

    return are_xarrays_broadcastable(y, x)


def broadcast_xarrays(x: DataArray, y: DataArray) -> Tuple[ndarray, ndarray, Tuple[str,...], Tuple[int,...]]:
    '''
    Determine how x and y should be broadcast. 

    Broadcastability is first checked by are_xarrays_broadcastable(x, y), if
    fails, an IndexError is raised.

    If x has more or equal dimensions than y, then the output dims will be
    conformed to x's, and vice versa.

    Return:
        A tuple of four objects:

            - ndarray of x
            - ndarray of y 
            (ready to be broadcast together)

            - output dimension names
            - output shape, e.g.

        x: dim (X, Y, Z, W) shape (1, 8, 32, 10)
        y: dim (X, W, Z) shape (9, 10, 1)

        => ndarray(1,8,32,10), ndarray(9,1,1,10), (X, Y, Z, W), (9, 8, 32, 10)
    '''
    if not are_xarrays_broadcastable(x, y): raise IndexError('xarrays cannot be broadcast')

    if len(x.dims) < len(y.dims):
        return broadcast_xarrays(y, x)

    
    _dims = cast(Tuple[str,...], x.dims)

    _x = x.__array__()
    _y = y.__array__() # has less axes than _x

    _shape = []

    _y_ax_map = dict()

    for dim in _dims:
        # axis_x will be the [axis] globally
        axis_x = cast(Tuple[int], x.get_axis_num((dim,)))[0]

        if dim in y.dims:
            axis_y = cast(Tuple[int], y.get_axis_num((dim,)))[0]
            _shape.append(max(x.shape[axis_x], y.shape[axis_y]))

            _y_ax_map[axis_x] = axis_y # map axis_y -> axis_x for axis permutation

        else: # there is no corresponding axis in y
            _y = _y[..., None] # add one axis
            _shape.append(x.shape[axis_x])

            _y_ax_map[axis_x] = len(_y.shape) - 1 # map (shape=1 new axis) -> axis_x

    _y_transpose_axes = tuple(_y_ax_map[_ax] for _ax in range(len(_y.shape)))
    _y = _y.transpose(_y_transpose_axes)

    _shape = tuple(_shape)
    return _x, _y, _dims, _shape



def broadcast_arrays(x: NDArray, y: NDArray) -> Tuple[ndarray, ndarray, Tuple[str,...]|None, Tuple[int,...]]:
    '''
    Try to broadcast two arrays.

    Case 1: x and y are both named (i.e. xarrays):

        returns broadcast_xarrays(x, y)

    Case 2: x and y are both unnamed (i.e. torch tensors or np ndarrays):

        use the default broadcasting logic for numpy arrays

    Case 3: the named tensor has more axes:
        
        the returned dimension names and shape are copied from the named tensor

    Caes 4: the unnamed tensor has more axes:

        e.g. x (64, 32, 32, 8), y (A=32, B=32, C=8)

        the returned dimension names and shape will be (dim0=64, A=32, B=32, C=8)
            
    '''
    if not are_arrays_broadcastable(x, y):
        raise IndexError('arrays cannot be broadcast')

    if isinstance(x, DataArray) and isinstance(y, DataArray):
        return broadcast_xarrays(x, y)



    _x = x.__array__()
    _y = y.__array__()


    shape = broadcast_shapes(x.shape, y.shape)

    if (not isinstance(x, DataArray)) and (not isinstance(y, DataArray)):
        return _x, _y, None, shape

    if isinstance(x, DataArray):
        _named = x 
    else:
        _named = y

    _named = cast(DataArray, _named)
    named_dims = cast(Tuple[str,...], _named.dims)

    if len(x.shape) > len(y.shape):
        _longer = x
    else:
        _longer = y

    if _named is _longer:
        dims = named_dims
    else:
        # named tensor is shorter
        dims = [f'dim_{i}' for i in range(len(_longer.shape))]
        dims[-len(named_dims):] = named_dims
        dims = tuple(dims)

    return _x, _y, dims, shape


    




from typing import Any, Protocol, Tuple, cast

from xtensors.typing import NDArray, DimLike
from xtensors.unify import get_axes, strip_dims

from xarray import DataArray

import numpy as np

'''
For reduction functions that only make sense when acting over a single axis
E.g. argmax
'''


class _np_single_reduction_func(Protocol):
    def __call__(self, a: NDArray, axis: int) -> DataArray: ...

class SingleDimReductionFunc(Protocol):
    def __call__(self, x: NDArray, dim: DimLike) -> DataArray: ...


def _reduction_factory(_np_func: _np_single_reduction_func) -> SingleDimReductionFunc:
    def _reduce(x: NDArray, dim: DimLike) -> DataArray:
        axis = get_axes(x, dim)[0]

        _y = _np_func(x.__array__(), axis=axis)

        dims = None
        if isinstance(x, DataArray):
            _dims = cast(Tuple[str,...], x.dims)
            dims = strip_dims(_dims, (axis,))

        return DataArray(_y, dims=dims)


    return _reduce


_argmax = _reduction_factory(np.argmax)
_argmin = _reduction_factory(np.argmin)

_nanargmax = _reduction_factory(np.nanargmax)
_nanargmin = _reduction_factory(np.nanargmin)



def _coord_reduc_factory(_func: SingleDimReductionFunc) -> SingleDimReductionFunc:

    def _reduce(x: NDArray, dim: DimLike) -> DataArray:
        '''
            Return the coordinate on [dim] that maximizes/minimizes x If dimension [dim] does
            not have coordinates, this is equivalent to (nan)argmax/argmin
        '''
        if not isinstance(x, DataArray):
            x_ = DataArray(x)
        else:
            x_ = x
        x_ = cast(DataArray, x)
            
        axis = get_axes(x_, dim)[0]
        args = _func(x_, dim)

        return x_.coords[x_.dims[axis]][args]
    return _reduce


_coordmax = _coord_reduc_factory(_argmax)
_coordmin = _coord_reduc_factory(_argmin)


_nancoordmax = _coord_reduc_factory(_nanargmax)
_nancoordmin = _coord_reduc_factory(_nanargmin)




from __future__ import annotations
from typing import Optional, overload
import numpy as np

from xtensors.functionals.base import DataArray, Functional
from xtensors import base
from xtensors.typing import NDArray, DimLike, DimsLike



class Reduction(Functional):
    '''
    Single-axis reduction fucntional
    '''
    def __init__(self, dim: DimLike) -> None:
        self.dim = dim
        self._reduce: base._reduc.ReductionFunc | base._reduc_single_dim.SingleDimReductionFunc
        self.name = 'UNIMPLEMENTED_REDUCTION'

    def __call__(self, x: NDArray) -> DataArray:
        return self._reduce(x, self.dim)


class Mean(Reduction):
    def __init__(self, dim: int | str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nanmean if nan else base.mean
        self.name = f'Mean({dim})'


class Sum(Reduction):
    def __init__(self, dim: int | str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nansum if nan else base.sum
        self.name = f'Sum({dim})'


class Std(Reduction):
    def __init__(self, dim: int | str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nanstd if nan else base.std
        self.name = f'Std({dim})'


class Max(Reduction):
    def __init__(self, dim: int | str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nanmax if nan else base.max
        self.name = f'Max({dim})'


class Min(Reduction):
    def __init__(self, dim: int | str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nanmin if nan else base.min
        self.name = f'Max({dim})'


class ArgMax(Reduction):
    def __init__(self, dim: int | str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nanargmax if nan else base.argmax
        self.name = f'ArgMax({dim})'


class ArgMin(Reduction):
    def __init__(self, dim: int | str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nanargmin if nan else base.argmin
        self.name = f'ArgMin({dim})'


class CoordMax(Reduction):
    def __init__(self, dim: int | str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nancoordmax if nan else base.coordmax
        self.name = f'CoordMax({dim})'


class CoordMin(Reduction):
    def __init__(self, dim: int | str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nancoordmin if nan else base.coordmin
        self.name = f'CoordMin({dim})'



from typing import Protocol
from numpy.typing import ArrayLike

from numpy import typing

import numpy as np
from xarray import DataArray
from xtensors.typing import NDArray
import torch


class _ufunc(Protocol):
    def __call__(self, __x1: np.ndarray, *args) -> np.ndarray: ...


class UFunc(Protocol):
    def __call__(self, x: NDArray) -> DataArray: ...



def _ufunc_factory(_np_func: _ufunc) -> UFunc:
    def _f(x: NDArray) -> DataArray:
        _x = np.array(x)
        _y = _np_func(_x)
        _dims = None
        
        if isinstance(x, DataArray):
            _dims = x.dims
        return DataArray(_y, dims=_dims)
    return _f


def _sigmoid(__x1: np.ndarray):

    return .5 * (1. + np.tanh(__x1/2))




sigmoid = _ufunc_factory(_sigmoid)

exp = _ufunc_factory(np.exp)
log = _ufunc_factory(np.log)
log10 = _ufunc_factory(np.log10)

cos = _ufunc_factory(np.cos)
sin = _ufunc_factory(np.sin)
tan = _ufunc_factory(np.tan)

cosh = _ufunc_factory(np.cosh)
sinh = _ufunc_factory(np.sinh)
tanh = _ufunc_factory(np.tanh)



from __future__ import annotations
from typing import Optional, Protocol, Tuple, cast

from xtensors.typing import NDArray
from xarray import DataArray

from ._broadcast import are_shapes_broadcastable, broadcast_xarrays, broadcast_arrays

from ..unify import get_coord

import xarray as xr
import warnings


class BinaryOperation(Protocol):
    def __call__(self, x: NDArray, y: NDArray) -> DataArray: ...


def _apply_operation(x: NDArray, y: NDArray, binop: str, rbinop: Optional[str]=None) -> NDArray:

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', 'invalid value encountered in true_divide')
        try:
            z = getattr(x, binop)(y)
            if z is NotImplemented: raise NotImplementedError

        except (AttributeError, NotImplementedError) as e:
            if rbinop is not None:
                z = getattr(y, rbinop)(x)
                if z is NotImplemented: raise NotImplementedError
            else:
                raise AttributeError(e) 
        
    return z


def _binop_factory(_bin_op: str, _rbin_op: str) -> BinaryOperation:
    def _op(x: NDArray, y: NDArray) -> DataArray:

        _x, _y, dims, _, = broadcast_arrays(x, y)
        _z = _apply_operation(_x, _y, _bin_op, _rbin_op)

        coords_map = dict()
        if dims is not None:
            for dimkey in dims:
                coord_x = get_coord(_x, dimkey)
                coord_y = get_coord(_y, dimkey)

                if coord_x is not None: coords_map[dimkey] = coord_x
                elif coord_y is not None: coords_map[dimkey] = coord_y

        return DataArray(_z, dims=dims, coords=coords_map)
    return _op

import numpy as np

_add = _binop_factory('__add__', '__radd__')
_divide = _binop_factory('__truediv__', '__rtruediv__')
_multiply = _binop_factory('__mul__', '__rmul__')

_greater = _binop_factory('__gt__', '__lt__')
_greater_equal = _binop_factory('__ge__', '__le__')

_less = _binop_factory('__lt__', '__gt__')
_less_equal = _binop_factory('__le__', '__ge__')

_equal = _binop_factory('__eq__', '__eq__')







from __future__ import annotations
from typing import cast, overload, Tuple, Protocol, Any

import numpy as np

from xarray import DataArray

from xtensors.typing import NDArray
from xtensors.typing import DimLike, DimsLike

from xtensors.unify import get_axes, strip_dims, get_coord



def diagonal(x: NDArray, dim1: DimLike, dim2: DimLike, dim_out: str) -> DataArray:
    '''
        Reduce a tensor x by taking the diagonal elements along [dim1] and
        [dim2]
    '''
    assert isinstance(dim1, (str, int, tuple))
    _x = x.__array__()

    axes1 =  get_axes(x, dim=dim1)
    axes2 =  get_axes(x, dim=dim2)

    assert len(axes1) == len(axes2) == 1

    axis1 = axes1[0]
    axis2 = axes2[0]

    y = np.diagonal(_x, axis1=axis1, axis2=axis2)

    if isinstance(x, DataArray):
        olddims = cast(Tuple[str], x.dims)
        newdims = list(strip_dims(olddims, (axis1, axis2)))

        coords_map = dict()
        for dimkey in newdims:
            coord = get_coord(x, dimkey)
            if coord is not None: coords_map[dimkey] = coord

        newdims.append(dim_out)
        
        
        return DataArray(y, dims=newdims, coords=coords_map)
    else:
        y_ = DataArray(y)
        y_ = y_.rename({y_.dims[-1]: dim_out})
        return y_




from .base import where,softmax, get_rank

from .base import (diagonal, 
        sum, std, mean, 
        max, min, all, any, 
        nanmean, nansum, nanstd)


from .base import (
        exp, sigmoid, log, log10,
        cos, sin, tan,
        cosh, sinh, tanh
        )


from .base import (
        are_shapes_broadcastable, broadcast_shapes,
        are_xarrays_broadcastable, broadcast_xarrays, 
        are_arrays_broadcastable, broadcast_arrays
        )

from .base import (
        argmax, argmin, nanargmax, nanargmin,
        coordmax, coordmin, nancoordmax, nancoordmin
        )

from .base import (
        add, divide, multiply,
        greater, greater_equal,
        less, less_equal,
        equal
        )


from .unify import (
        rename, 
        strip_dims, get_axis_from_dim, get_axes, 
        new_axes, new_axis)




#from .functionals.base import Pipe, Functional
